#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#define _USE_MATH_DEFINES // for C++
#include <cmath>
#include <time.h>
#include <ctime>
#include "idx.h"
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/variate_generator.hpp>

boost::mt19937 seed(std::time(0));

int random(int limit){
	boost::uniform_int<> dist(0, limit-1);
	boost::variate_generator<boost::mt19937&, boost::uniform_int<> > getNumber(seed, dist);
	return getNumber();
}

int myroundf(float x){
	float f=floorf(x);
	return x-f>=0.5f ? f+1.0f : f;
}

int myround(double x){
	double f=floor(x);
	return x-f>=0.5 ? f+1.0 : f;
}

IDXHandler::IDXHandler(int _bufferSize){
	nImages = 0;
	width  = 0;
	height = 0;
	buffer = NULL;
	labels = NULL;
	indexes = NULL;
	bufferDoubles = NULL;
	bufferSize=_bufferSize;
}

IDXHandler::~IDXHandler(){
}

//reverses a 32 bit unsigned integer 
inline void 
IDXHandler::reverse(unsigned  &sig){ if(reverted) sig=(sig<<24) | ((sig<<8)&0x00FF0000) | ((sig>>8)&0x0000FF00) | (sig>>24); }

void 
IDXHandler::loadIDXimages(const char *filename){
	FILE *file=NULL;
	clock_t t1,t2;
	t1 = clock(); 
	printf("Loading image file %s ...",filename);
	strcpy(this->filename,filename);

	if ((file=fopen(filename,"rb")) == NULL){
		printf("Error: can not open file %s\n",filename);
		return;
	}

	/* read the magic number */
	fread(&signature,sizeof(signature),1,file); printf("Signature: %X\n",signature);
	reverted = ((signature & 0x0FF) == 0); reverse(signature);
	printf("Signature: %X\n",signature);
	datatype = (signature>>8) & 0x0FF;
	dimmensions = signature & 0x0FF;
	if(datatype!=8){
		printf("Error in %s: Data(%ud) in %s is not byte(8).\n\7",__FUNCTION__,filename,datatype); return;
	}
	if(dimmensions!=3){
		printf("Error in %s (file: %s): Data has %ud dimmensions. Only 3 are alowed.\n\7",__FUNCTION__,filename,dimmensions); return;
	}
	
	/* read the number of images, then the size of an image */
	unsigned int x;
	fread(&x,sizeof(nImages),1,file); reverse(x); 
	if(nImages!=0){
		if(nImages>=0 && nImages<=x) 
			printf("Loading %d images out of %d.\7\n",nImages,x);
		else{ printf("Cannot read %d images. The file contains %d.\7\n", nImages, x); exit(1);}
	}
	else nImages=x;

	fread(&x,sizeof(x),1,file); reverse(x); height=x;
	fread(&x,sizeof(x),1,file); reverse(x); width=x;
	
	samplesize=height*width;
	headersize = 4*(1+dimmensions);

	/* initialize the indexes */

	if( (indexes = (int  *)malloc(nImages * sizeof(int))) == NULL) {printf("Error allocating space for indexes[] in %s. \n\7",__FUNCTION__); exit(1);}
	for (int  i=0; i<nImages; i++) 
		indexes[i] = i;

	buffer = (unsigned char *)malloc(nImages*samplesize*sizeof(unsigned char));
	if (buffer == NULL) {
		printf("Error: cannot allocate %d bytes for buffer in %s\n\7", nImages*samplesize*sizeof(unsigned char),__FUNCTION__);
		printf("Switching to streaming ...\n\7");
		exit(1);
	}
	else{
		/* read all images at once */
		fread(buffer,sizeof(unsigned char), nImages*samplesize, file);
		fclose(file);
	}
				
	t2 = clock();
	printf("done in %6.4lf s\n",(double)(t2-t1)/CLK_TCK);
}

void 
IDXHandler::saveIDXimages(const char *filename){
	FILE *f;
	if ((f=fopen(filename,"wb")) == NULL){
		printf("Error: can not open file %s\n",filename);
		return;
	}

	/* write the magic number */
	unsigned  signature = 0x8030000; //...803 caci am nevoie sa salvez doar setul de antrenare
	reverse(signature);	fwrite(&signature,sizeof(signature),1,f); 

	/* write header */
	unsigned  r_nImages = nImages;
	unsigned  r_height  = height;
	unsigned  r_width   = width;
	reverse(r_nImages); fwrite(&r_nImages,sizeof(r_nImages),1,f); 
	reverse(r_height);  fwrite(&r_height,sizeof(r_height),1,f); 
	reverse(r_width);   fwrite(&r_width,sizeof(r_width),1,f); 

	/* write all images at once */
	fwrite(buffer, sizeof(unsigned char), nImages*width*height,f);	

	fclose(f);
}

int fcmp_int(const void *a, const void *b){	return *(int *)a - *(int *)b; }

void 
IDXHandler::loadIDXlabels(const char *filename){
	FILE *f;
	if ((f=fopen(filename,"rb")) == NULL){ fprintf(stderr, "Error: can not open file %s\n\7",filename); exit(1); }
	
	/* read the magic number */
	unsigned  signature;
	fread(&signature,sizeof(signature),1,f); reverse(signature);
	if((signature&0xff)!=1){printf("Dimensionality of the label file should be 1, not %d!\n\7",signature&0xff);exit(1);}
	if(((signature>>8)&0xff)!=8 && ((signature>>8)&0xff)!=16 && ((signature>>8)&0xff)!=32){printf("Labels should be 8, 16 or 32 bit, not %d!\n\7",(signature>>8)&0xff); exit(1);}

	/* read the number of labels */
	unsigned  nr; fread(&nr,sizeof(nr),1,f); reverse(nr);
 
	/* the number of labels and the number of images must "match" */
	if (nr != nImages){
		printf("Error: The IDX Label File contains a different # labels (%d) than the # of images (%d)!\n",nr, nImages);
		exit(1);
	}
	
	/* read all labels */
	labels = (int *)malloc(nImages * sizeof(int));
	unsigned char luc;
	for(int i=0;i<nImages;i++){
		fread(&luc,sizeof(unsigned char),1,f);
		labels[i]=luc;
	}
	
	fclose(f);

	//counting the number of different classes and building the map (the final map can have labels in a not ascending order!)
	n_labels=0;
	for(int i=0; i<nImages; i++){
		int found=0;
		for(int j=0; j<n_labels; j++)
			if(label_map[j]==labels[i]){found=1; break;}
		if(!found){label_map[n_labels]=labels[i]; n_labels++;}
	}
	qsort(label_map,n_labels,sizeof(int),fcmp_int);

	//computing samples per class
	for(int i=0; i<n_labels; i++) samplesPerClass[i]=0;

	for(int i=0; i<nImages; i++)
		samplesPerClass[getLabel(i)]++;
}

int 
IDXHandler::getLabel(int idx){
	int i,label,index = indexes[idx];
	
	label=labels[index];

	for(i=0;i<n_labels;i++)
		if(label==label_map[i])break;
	return i;
}

void 
IDXHandler::unloadIDX(){
	if (buffer != NULL){
		nImages = 0;
		width = 0;
		height = 0;
		free(buffer);
		buffer = NULL;
	}
	if (bufferDoubles!=NULL) {free(bufferDoubles); bufferDoubles=NULL;}

	if (labels != NULL) {free(labels); labels=NULL;}

	if (indexes != NULL) {free(indexes); indexes=NULL;}
}

void 
IDXHandler::scramble(){
	if (nImages == 0) {	printf("Error: Can't scramble. No images were loaded.\n"); return; }
	int r1,r2;
	for (int i=0; i<nImages; i++){
		r1 = random(nImages);
		r2 = random(nImages);
		switchidx(r1,r2);
	}
}

inline void 
IDXHandler::switchidx(int i1, int i2){
	int  temp; temp = indexes[i1]; indexes[i1] = indexes[i2]; indexes[i2] = temp;
}

double *
IDXHandler::getImageAsDoubles(int idx){
	double *temp = (double *)malloc(samplesize*sizeof(double));

	for (int i=0; i<samplesize; i++)
		temp[i] = buffer[indexes[idx]*samplesize + i] / 127.5 - 1.0;
	
	return temp;
}

void 
IDXHandler::saveErrorImages(char *filename, Error *errorvector, int size){
	FILE *fout;
	if ((fout=fopen(filename,"wb")) == NULL){
		printf("Error: can not open file %s\n",filename);
		return;
	}

	/* write the magic number */
	unsigned long sig=(signature&0xFFFFFF00u);
	sig+= 0x03u;
	fwrite(&sig,sizeof(sig),1,fout); 
	
	/* write the number of images, then the size of an image */
	unsigned  r = size;	fwrite(&r,sizeof(r),1,fout);
	r = height;	fwrite(&r,sizeof(r),1,fout);
	r = width;	fwrite(&r,sizeof(r),1,fout); 

	__int64 imgOffset;

	unsigned char *tempbuffer = (unsigned char *)malloc(samplesize*sizeof(unsigned char));
	if (tempbuffer == NULL) {
		printf("Error: malloc tempbuffer IDX in function %s, file %s, line %d.\n",__FUNCTION__,__FILE__,__LINE__);
		return;
	}

	for (int k=0; k<size; k++){
		imgOffset = (__int64)errorvector[k].idx*(__int64)samplesize;
		for (int i=0; i<samplesize; i++)
			tempbuffer[i] = buffer[imgOffset + i];
		// write the image
		fwrite(tempbuffer,sizeof(unsigned char),samplesize,fout);	
	}

	fclose(fout);
	free(tempbuffer);
}

void 
IDXHandler::allocateBuffer(void){
	bufferDoubles = (double *)malloc(bufferSize*samplesize*sizeof(double));
	if(bufferDoubles==NULL){printf("Allocation error for bufferDoubles in %s\n\7",__FUNCTION__);exit(1);}
}

void 
IDXHandler::collectData(int startIdx){
	for(int i=startIdx; i<startIdx+bufferSize && i<nImages; i++)
		for(int j=0; j<samplesize; j++)
			bufferDoubles[(i-startIdx)*samplesize+j] = buffer[indexes[i]*samplesize + j] / 127.5 - 1.0;
}